#include "id_sd.h"
//#include "fmopl.h"
#include "dbopl.h"
#include <SDL.h>
#include <string>

typedef Bit8u byte;
typedef Bit16u word;
typedef Bit32s fixed;
typedef Bit32u longword;
typedef Bit8s boolean;
typedef void * memptr;

//      Sequencer variables
static  longword                alTimeCount;
static  volatile boolean        sqActive;
static  word                   *sqHack;
static  word                   *sqHackPtr;
static  int                     sqHackLen;
static  int                     sqHackSeqLen;
static  longword                sqHackTime;
static  boolean                 isKMF;

int sdl_MixerRate;
boolean rate_is_set;

int numreadysamples = 0;
int samplesPerMusicTick;

static Chip opl_chip;

static Bit32s *mix_buffer = NULL;

inline void OPLUpdate(Sint16 *buffer, int length)
{
	Bit32s mix, *buff;

	Chip__GenerateBlock2(&opl_chip, length, mix_buffer );

	// Mix into the destination buffer, doubling up into stereo.
	buff = mix_buffer;
	while(length--)
	{
		mix = *buff++;
		mix <<= 2;		// increase volume a bit

		if(mix > 32767)
			mix = 32767;
		else if(mix < -32768)
			mix = -32768;

		*buffer++ = (Sint16)mix;
		*buffer++ = (Sint16)mix;
	}
}


void SDL_IMFMusicPlayer(void *udata, Uint8 *stream, int len)
{
	int stereolen = len>>1;
	int sampleslen = stereolen>>1;
	Sint16 *stream16 = (Sint16 *) (void *) stream;    // expect correct alignment

	while(1)
	{
		if (numreadysamples)
		{
			if(numreadysamples<sampleslen)
			{
				OPLUpdate(stream16, numreadysamples);
				stream16 += numreadysamples*2;
				sampleslen -= numreadysamples;
			}
			else
			{
				OPLUpdate(stream16, sampleslen);
				numreadysamples -= sampleslen;
				return;
			}
		}
		if (sqActive)
		{
			if (isKMF)
			{
				Uint8 count, ticks;
				while (sqHackLen > 0 && sqHackTime <= alTimeCount)
				{
					count = ((byte *)sqHackPtr)[0];
					ticks = ((byte *)sqHackPtr)[1];
					sqHackPtr++;
					sqHackTime = alTimeCount + ticks;
					sqHackLen -= (count+1) << 1;
					while (count--)
					{
						Chip__WriteReg(&opl_chip, *(byte *) sqHackPtr, *(((byte *) sqHackPtr)+1) );
						sqHackPtr++;
					}
				}
			}
			else
			{
				do
				{
					if(sqHackTime > alTimeCount) break;
					sqHackTime = alTimeCount + SDL_SwapLE16(*(sqHackPtr+1));
					Chip__WriteReg(&opl_chip, *(byte *) sqHackPtr, *(((byte *) sqHackPtr)+1) );
					sqHackPtr += 2;
					sqHackLen -= 4;
				}
				while(sqHackLen>0);
			}
			alTimeCount++;
			if(!sqHackLen)
			{
				sqHackPtr = sqHack;
				sqHackLen = sqHackSeqLen;
				sqHackTime = 0;
				alTimeCount = 0;
			}
		}
		numreadysamples = samplesPerMusicTick;
	}

	//SDL_MixAudio(stream, 0, len, SDL_MIX_MAXVOLUME);	//we don't need to mix it with nothing! --K1n9_Duk3
}

///////////////////////////////////////////////////////////////////////////
//
//      SD_Startup() - starts up the Sound Mgr
//              Detects all additional sound hardware and installs my ISR
//
///////////////////////////////////////////////////////////////////////////
void
SD_Startup(int imf_clock_rate, int mixer_rate, int opl_rate)
{
    // Open the audio device
    SDL_AudioSpec *desired; //, *obtained;	//not really required --K1n9_Duk3

    // initialize SDL
    if(SDL_Init( SDL_INIT_AUDIO ) < 0)
    {
        printf("Unable to init SDL: %s\n", SDL_GetError());
        exit(1);
    }
    atexit(SDL_Quit);
	
	// Allocate a desired SDL_AudioSpec
    desired = (SDL_AudioSpec*) malloc(sizeof(SDL_AudioSpec));
    //obtained = (SDL_AudioSpec*) malloc(sizeof(SDL_AudioSpec)); //not really required --K1n9_Duk3
    desired->freq=mixer_rate;
    desired->format=AUDIO_S16SYS;
    desired->channels=2;
    desired->samples = (mixer_rate * 2048) / 44100;
    desired->callback=SDL_IMFMusicPlayer;
    desired->userdata=NULL;

    // Open the audio device
    if ( SDL_OpenAudio(desired, NULL) < 0 ){	//forces SDL to convert audio to the desired format --K1n9_Duk3
      fprintf(stderr, "Couldn't open audio: %s\n", SDL_GetError());
      exit(1);
    }

    // desired spec is no longer needed
    free(desired);

    // Init music

    sdl_MixerRate = mixer_rate;	// K1n9_Duk3 addition (so that KMF files can adjust playback rate)
    if (imf_clock_rate)
    {
        rate_is_set = true;
        samplesPerMusicTick = mixer_rate / imf_clock_rate;    // SDL_t0FastAsmService played at imf_clock_rate Hz
    }
    else
    {
        rate_is_set = false;
        samplesPerMusicTick = mixer_rate / 560;
    }

    mix_buffer = (Bit32s *) malloc(mixer_rate * sizeof(Bit32s));

    DBOPL_InitTables();
    Chip__Chip(&opl_chip);
    Chip__Setup(&opl_chip, mixer_rate);

    Chip__WriteReg(&opl_chip, 1, 0x20);	//set WSE=1

    SDL_PauseAudio(0);
    alTimeCount = 0;
}

int IMF_IsChunk(FILE* in)
{
	Uint16 chunksize, buff, i=42;
	Uint32 sum1=0, sum2=0;

	if(!in)
		return 0;

	if (!feof(in))
	{
		fread(&chunksize, sizeof(chunksize), 1, in);
		while(!feof(in) && i)
		{
			fread(&buff, sizeof(buff), 1, in);
			sum1 += SDL_SwapLE16(buff);
			fread(&buff, sizeof(buff), 1, in);
			sum2 += SDL_SwapLE16(buff);
			i--;
		}
		fseek(in, 0, SEEK_SET);
		return (sum1 > sum2);
	}
	return 0;
}
///////////////////////////////////////////////////////////////////////////
//
//      SD_StartMusic() - starts playing the music pointed to
//
///////////////////////////////////////////////////////////////////////////
void
SD_StartMusic(const std::string& filename)
{
	Uint32 size = 0;
	Uint16 size16;

	// Load the IMF File here!
	FILE *fp=fopen(filename.c_str(), "rb");
	if (fp)
	{
		fread(&size, sizeof(size), 1, fp);
		if (SDL_SwapLE32(size) == 0x1A464D4B)
		{
			Uint16 rate;
			fread(&rate,sizeof(rate),1,fp);
			rate = SDL_SwapLE16(rate);
			if (rate != 0 && !rate_is_set)
				samplesPerMusicTick = sdl_MixerRate / rate;
			fread(&size16,sizeof(size16),1,fp);
			size = SDL_SwapLE16(size16);
			printf("Song is in KMF format. %i Hz, %i bytes\n\n", (int)rate, (int)size);
			isKMF = 1;
		}
		else
		{
			isKMF = 0;
			fseek(fp, 0, SEEK_SET);
			if (IMF_IsChunk(fp))
			{
				fread(&size16,sizeof(Uint16),1,fp);
				size = SDL_SwapLE16(size16);
				printf("Song is a ripped AudioT chunk. %i bytes\n\n", (int)size);
			}
			else
			{
				fseek(fp, 0L, SEEK_END);
				size = ftell(fp);
				fseek(fp, 0L, SEEK_SET);
				printf("Song is a plain IMF file. %i bytes\n\n", (int)size);
			}
		}

		byte* IMFBuffer = (byte*) malloc(size);

		fread(IMFBuffer,sizeof(byte),size,fp);

		fclose(fp);

		sqHack = (word *)(void *) IMFBuffer;

		sqHackLen = sqHackSeqLen = size;
		sqHackPtr = sqHack;
		sqHackTime = 0;
		alTimeCount = 0;

		sqActive = true;
	} else {
		printf("ERROR: could not open %s\n", filename.c_str());
	}
}

///////////////////////////////////////////////////////////////////////////
//
//      SD_Shutdown() - shuts down the Sound Mgr
//              Removes sound ISR and turns off whatever sound hardware was active
//
///////////////////////////////////////////////////////////////////////////
void
SD_Shutdown(void)
{
	// Prevent the audio callback from accessing the buffers:
	sqActive = false;
	SDL_PauseAudio(1);
	SDL_Delay(50);

	if(mix_buffer)
		free(mix_buffer);

	if(sqHack)
		free(sqHack);
}
